chiark / gitweb /
help: use pager
[topgit.git] / tg.sh
diff --git a/tg.sh b/tg.sh
index 1fe01155a27bc7e1bfff1de04d68e1dd62ab01ad..4185ba2f3baac4541c204d2aa704d7bbd1e49d52 100644 (file)
--- a/tg.sh
+++ b/tg.sh
@@ -3,6 +3,7 @@
 # (c) Petr Baudis <pasky@suse.cz>  2008
 # GPLv2
 
+TG_VERSION=0.7
 
 ## Auxiliary functions
 
@@ -17,6 +18,27 @@ die()
        exit 1
 }
 
+# cat_file "topic:file"
+# Like `git cat-file blob $1`, but topics '(i)' and '(w)' means index and worktree
+cat_file()
+{
+       arg="$1"
+       case "$arg" in
+       '(w):'*)
+               arg=$(echo "$arg" | tail --bytes=+5)
+               cat "$arg"
+               return
+               ;;
+       '(i):'*)
+               # ':file' means cat from index
+               arg=$(echo "$arg" | tail --bytes=+5)
+               git cat-file blob ":$arg"
+               ;;
+       *)
+               git cat-file blob "$arg"
+       esac
+}
+
 # setup_hook NAME
 setup_hook()
 {
@@ -63,8 +85,8 @@ measure_branch()
        _bname="$1"; _base="$2"
        [ -n "$_base" ] || _base="refs/top-bases/$_bname"
        # The caller should've verified $name is valid
-       _commits="$(git rev-list "$_bname" ^"$_base" | wc -l)"
-       _nmcommits="$(git rev-list --no-merges "$_bname" ^"$_base" | wc -l)"
+       _commits="$(git rev-list "$_bname" ^"$_base" -- | wc -l)"
+       _nmcommits="$(git rev-list --no-merges "$_bname" ^"$_base" -- | wc -l)"
        if [ $_commits -gt 1 ]; then
                _suffix="commits"
        else
@@ -77,7 +99,7 @@ measure_branch()
 # Whether B1 is a superset of B2.
 branch_contains()
 {
-       [ -z "$(git rev-list ^"$1" "$2" --)" ]
+       [ -z "$(git rev-list --max-count=1 ^"$1" "$2" --)" ]
 }
 
 # ref_exists REF
@@ -94,6 +116,16 @@ has_remote()
        [ -n "$base_remote" ] && ref_exists "remotes/$base_remote/$1"
 }
 
+branch_annihilated()
+{
+       _name="$1";
+
+       # use the merge base in case the base is ahead.
+       mb="$(git merge-base "refs/top-bases/$_name" "$_name")";
+
+       test "$(git rev-parse "$mb^{tree}")" = "$(git rev-parse "$_name^{tree}")";
+}
+
 # recurse_deps CMD NAME [BRANCHPATH...]
 # Recursively eval CMD on all dependencies of NAME.
 # CMD can refer to $_name for queried branch name,
@@ -104,6 +136,7 @@ has_remote()
 # of the whole function.
 # If recurse_deps() hits missing dependencies, it will append
 # them to space-separated $missing_deps list and skip them.
+# remote dependencies are processed if no_remotes is unset.
 recurse_deps()
 {
        _cmd="$1"; shift
@@ -111,12 +144,17 @@ recurse_deps()
        _depchain="$*"
 
        _depsfile="$(mktemp -t tg-depsfile.XXXXXX)"
-       # Check also our base against remote base. Checking our head
-       # against remote head has to be done in the helper.
-       if has_remote "top-bases/$_name"; then
+       # If no_remotes is unset check also our base against remote base.
+       # Checking our head against remote head has to be done in the helper.
+       if test -z "$no_remotes" && has_remote "top-bases/$_name"; then
                echo "refs/remotes/$base_remote/top-bases/$_name" >>"$_depsfile"
        fi
-       git cat-file blob "$_name:.topdeps" >>"$_depsfile"
+
+       # if the branch was annihilated, there exists no .topdeps file
+       if ! branch_annihilated "$_name"; then
+               #TODO: handle nonexisting .topdeps?
+               git cat-file blob "$_name:.topdeps" >>"$_depsfile";
+       fi;
 
        _ret=0
        while read _dep; do
@@ -189,7 +227,7 @@ needs_update()
 # branch_empty NAME
 branch_empty()
 {
-       [ -z "$(git diff-tree "refs/top-bases/$1" "$1" | fgrep -v "     .top")" ]
+       [ -z "$(git diff-tree "refs/top-bases/$1" "$1" -- | fgrep -v "  .top")" ]
 }
 
 # switch_to_base NAME [SEED]
@@ -228,10 +266,11 @@ do_help()
                        sep="|"
                done
 
-               echo "TopGit v0.5 - A different patch queue manager"
+               echo "TopGit v$TG_VERSION - A different patch queue manager"
                echo "Usage: tg [-r REMOTE] ($cmds|help) ..."
        elif [ -r "@cmddir@"/tg-$1 ] ; then
-               @cmddir@/tg-$1 -h || :
+               setup_pager
+               @cmddir@/tg-$1 -h 2>&1 || :
                echo
                if [ -r "@sharedir@/tg-$1.txt" ] ; then
                        cat "@sharedir@/tg-$1.txt"
@@ -243,6 +282,44 @@ do_help()
        fi
 }
 
+## Pager stuff
+
+# isatty FD
+isatty()
+{
+       test -t $1
+}
+
+# setup_pager
+# Spawn pager process and redirect the rest of our output to it
+setup_pager()
+{
+       isatty 1 || return 0
+
+       # TG_PAGER = GIT_PAGER | PAGER | less
+       # NOTE: GIT_PAGER='' is significant
+       TG_PAGER=${GIT_PAGER-${PAGER-less}}
+
+       [ -z "$TG_PAGER"  -o  "$TG_PAGER" = "cat" ]  && return 0
+
+
+       # now spawn pager
+       export LESS=${LESS:-FRSX}       # as in pager.c:pager_preexec()
+
+       _pager_fifo_dir="$(mktemp -t -d tg-pager-fifo.XXXXXX)"
+       _pager_fifo="$_pager_fifo_dir/0"
+       mkfifo -m 600 "$_pager_fifo"
+
+       "$TG_PAGER" < "$_pager_fifo" &
+       exec > "$_pager_fifo"           # dup2(pager_fifo.in, 1)
+
+       # this is needed so e.g. `git diff` will still colorize it's output if
+       # requested in ~/.gitconfig with color.diff=auto
+       export GIT_PAGER_IN_USE=1
+
+       # atexit(close(1); wait pager)
+       trap "exec >&-; rm \"$_pager_fifo\"; rmdir \"$_pager_fifo_dir\"; wait" EXIT
+}
 
 ## Startup
 
@@ -265,8 +342,9 @@ setup_hook "pre-commit"
 ## Dispatch
 
 # We were sourced from another script for our utility functions;
-# this is set by hooks.
-[ -z "$tg__include" ] || return 0
+# this is set by hooks.  Skip the rest of the file.  A simple return doesn't
+# work as expected in every shell.  See http://bugs.debian.org/516188
+if [ -z "$tg__include" ]; then
 
 if [ "$1" = "-r" ]; then
        shift
@@ -298,3 +376,7 @@ help|--help|-h)
        }
        . "@cmddir@"/tg-$cmd;;
 esac
+
+fi
+
+# vim:noet